/*========================================================*/
/*                                                        */
/*  contlim.c           version 1.8.1   2007.02.22        */
/*                                                        */
/*  Copyright (C) 2007 by Wojtek Pych, CAMK PAN           */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*                                                        */
/*  Calculate useful range of wavelength for echelle      */
/*  apertures taking continuum fit information.           */
/*  Add user defined limits.                              */
/*                                                        */
/*========================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pfitsio.h"

#define LINELEN 200

  typedef struct PARS
  {
    short vopt;		// verbose level
    int   ap1;		// number of the first aperture
                        // ignored if ECORD# found in the header
    float countratio,	// max/min count ratio accepted
          countmin,	// minimum cont accepted
          difminratio,	// minimum differencial
          difmaxratio;	// maximum differencial
  } PARS;

void  usage(void);
PARS  readpar(void);
int   wat2(PARS, int, char **, int, int, double *, double *);
int   countlim(PARS, int, int, float *, int *, int *);
int   difmax(PARS, float *, int, int);
int   difmin(PARS, float *, int, int);

/*--------------------------------------------------------*/
void usage()
{
  printf("\n\t USAGE: contlim <input_file>\n\n");
  printf("input_file: FITS file with the fit of continuum\n\n");
  exit(EXIT_FAILURE);
}
/*--------------------------------------------------------*/
PARS readpar()
{
        char  str[LINELEN];
        FILE  *inf;
        PARS  lpar;

  if ((inf=fopen("contlim.par", "r")) == NULL)
  { perror("\n\t contlim.par"); exit(EXIT_FAILURE); }

/** number of the first aperture **/
  fgets(str, LINELEN, inf);
  if (feof(inf) || !strncasecmp(str, "end", 3))
  { printf("\n\tERROR! Premature end of 'contlim.par'\n"); exit(EXIT_FAILURE); }
  sscanf(str, "%d", &lpar.ap1);

/** countratio **/
  fgets(str, LINELEN, inf);
  if (feof(inf) || !strncasecmp(str, "end", 3))
  { printf("\n\tERROR! Premature end of 'contlim.par'\n"); exit(EXIT_FAILURE); }
  sscanf(str, "%f", &lpar.countratio);

/** countmin **/
  fgets(str, LINELEN, inf);
  if (feof(inf) || !strncasecmp(str, "end", 3))
  { printf("\n\tERROR! Premature end of 'contlim.par'\n"); exit(EXIT_FAILURE); }
  sscanf(str, "%f", &lpar.countmin);

/** difminratio **/
  fgets(str, LINELEN, inf);
  if (feof(inf) || !strncasecmp(str, "end", 3))
  { printf("\n\tERROR! Premature end of 'contlim.par'\n"); exit(EXIT_FAILURE); }
  sscanf(str, "%f", &lpar.difminratio);

/** difmaxratio **/
  fgets(str, LINELEN, inf);
  if (feof(inf) || !strncasecmp(str, "end", 3))
  { printf("\n\tERROR! Premature end of 'contlim.par'\n"); exit(EXIT_FAILURE); }
  sscanf(str, "%f", &lpar.difmaxratio);

/** vopt **/
  fgets(str, LINELEN, inf);
  if (feof(inf) || !strncasecmp(str, "end", 3))
  { printf("\n\tERROR! Premature end of 'contlim.par'\n"); exit(EXIT_FAILURE); }
  sscanf(str, "%hd", &lpar.vopt);

  fclose(inf);

  if (lpar.vopt > 0)
  {
    printf("\tReading contlim.par:\n");
    printf("\t--------------------\n");
    printf("number of the 1st aperture = %d\n", lpar.ap1);
    printf("minimum/maximum count ratio= %f\n", lpar.countratio);
    printf("minimum count=               %f\n", lpar.countmin);
    printf("minimum differential ratio=  %f\n", lpar.difminratio);
    printf("maximum differential ratio=  %f\n", lpar.difmaxratio);
    printf("verbosity level=             %d\n", lpar.vopt);
  }

  return(lpar);
}
/*--------------------------------------------------------*/
int wat2(PARS par, int hsize, char **header, int naxis1, int naxis2,
         double *w1, double *dw)
{
        char    fl[80],
                *wat;
        int     i;

  if ((wat=(char *)calloc(1, sizeof(char))) == NULL)
  { perror("\n\t wat2: calloc(wat)"); exit(EXIT_FAILURE); }

  for (i=0; i<hsize; i++)
  {
    strncpy(fl, header[i], 4);
    fl[4]='\0';
    if (strcmp(fl, "WAT2") == 0)
    {
      strcpy(fl, strchr(header[i], '\'')+1);
      *strrchr(fl, '\'')='\0';

      if (!(wat=(char *)realloc(wat, (strlen(wat)+strlen(fl)+1)*sizeof(char))))
      { perror("\n\tERROR! wat2: realloc(wat)"); exit(EXIT_FAILURE); }

      strcat(wat, fl);
    }
  }

  if (par.vopt > 1)  printf("%s\n", wat);

  wat=strstr(wat, "spec1");

  for (i=0; ((wat=strstr(wat, "spec"))); i++)
  {
    if (i == naxis2)
    {
      printf("\n\tERROR! wat2: too much information in 'WAT2'\n");
      exit(EXIT_FAILURE);
    }

    sscanf(wat, "%s %s %s %s %s %lf %lf", fl, fl, fl, fl, fl, &w1[i], &dw[i]);
    if (par.vopt)  printf("%f  %f\n", w1[i], dw[i]);
    wat++;
  }

  free(wat);

  if (i < naxis2)
  {
    printf("\n\tERROR! wat2: not enough information in 'WAT2'\n");
    exit(EXIT_FAILURE);
  }

  return(0);
}
/*--------------------------------------------------------*/
int countlim(PARS par, int n, int ap, float *apdata, int *ic1, int *ic2)
{
        int   i,
              imax;		// number of the pixel with maxmimum count
        float apdatamax,	// peak of the spectrum in the aperture
              lim;

/* top of the spectrum in the aperture */
  imax=0;
  apdatamax=apdata[0];
  for (i=1; i<n; i++)
  {
    if (apdata[i] > apdatamax)
    {
      apdatamax=apdata[i];
      imax=i;
    }
  }

  if ((imax == 0) || (imax == n-1))
  {
    fprintf(stderr, "aperture %d: peak on the edge, imax= %d\n", ap, imax);
    return(-1);
  }

/* minimum acceptable counts */
  lim=apdatamax/par.countratio;
  if (lim < par.countmin) lim=par.countmin;

  for (i=imax; i>0; i--) if (apdata[i] < lim) break;
  *ic1=i;

  for (i=imax; i<n; i++) if (apdata[i] < lim) break;
  if (i == n) i--;
  *ic2=i;

  if (*ic1 == *ic2)
  {
    fprintf(stderr, "aperture %d: counts too low, max= %g\n", ap, apdatamax);
    return(-1);
  }

  if (par.vopt > 1)
  {
    printf("maximum count(%d)= %.1f\n", imax, apdatamax);
    printf("minimum accepted count= %.1f\n", lim);
  }

  return(imax);
}
/*--------------------------------------------------------*/
int difmax(PARS par, float *apdata, int ic1, int imax)
{
        int   i,
              idmax,
              id1;
        float dif,	// local differencial
              dmax,	// maximum differencial
              lim;	// differencial limit

/* maximum differencial */

  dmax=apdata[imax]-apdata[imax-1];
  idmax=imax;

  for (i=imax-1; i>ic1; i--)
  {
    dif=apdata[i]-apdata[i-1];
    if (dif > dmax)
    {
      dmax=dif;
      idmax=i;
    }
  }

/* differencial limit */
  lim=dmax/par.difmaxratio;

  if (par.vopt > 1)
  {
    printf("maximum differential(%d - %d)= %G\n", idmax, idmax-1, dmax);
    printf("minimum accepted positive differential= %G\n", lim);
  }

  for (i=idmax; i>ic1; i--) if ((apdata[i]-apdata[i-1]) < lim) break;
  id1=i;  

  return(id1);
}
/*--------------------------------------------------------*/
int difmin(PARS par, float *apdata, int ic2, int imax)
{
        int   i,
              idmin,
              id2;
        float dif,	// local differencial
              dmin,	// minimum differencial
              lim;	// differencial limit

/* minimum differencial */

  dmin=apdata[imax+1]-apdata[imax];
  idmin=imax+1;

  for (i=imax+2; i<ic2; i++)
  {
    dif=apdata[i]-apdata[i-1];
    if (dif < dmin)
    {
      dmin=dif;
      idmin=i;
    }
  }

/* differencial limit */
  lim=dmin/par.difminratio;

  for (i=idmin; i<=ic2; i++)
    if ((apdata[i]-apdata[i-1]) > lim) break;
  if (i == ic2+1) i--;
  id2=i;

  if (par.vopt > 1)
  {
    printf("minimum negative differential(%d - %d)= %G\n",
            idmin, idmin-1, dmin);
    printf("maximum accepted negative differential= %G\n", lim);
  }

  return(id2);
}
/*--------------------------------------------------------*/
int main(int argc, char *argv[])
{
        char    iname[LINELEN],
                **header,
                ecor[KEYWORD_SIZE],
                val[VALUE_SIZE],
                wstr1[20], wstr2[20];
        int     i, j,
                hsize,
                naxis1, naxis2,
                *ap,
                imax,
                ic1, ic2,
                id1, id2;
        float   **data,
                *apdata;
        double  *w1, *dw;
        PARS    par;

  if (argc == 2)  strcpy(iname, argv[1]);
  else            usage();

  par=readpar();

  data=read_FITS_2Dfile(iname, 's', &hsize, &header, &naxis1, &naxis2);
  if (par.vopt) printf("\n%s: %d x %d\n\n", iname, naxis1, naxis2);

  if ((w1=(double *)calloc(naxis2, sizeof(double))) == NULL)
  { perror("\n\t calloc(w1)"); exit(EXIT_FAILURE); }

  if ((dw=(double *)calloc(naxis2, sizeof(double))) == NULL)
  { perror("\n\t calloc(dw)"); exit(EXIT_FAILURE); }

  wat2(par, hsize, header, naxis1, naxis2, w1, dw);

  if ((apdata=(float *)calloc(naxis1, sizeof(float))) == NULL)
  { perror("\n\t calloc(apdata)"); exit(EXIT_FAILURE); }

  if ((ap=(int *)calloc(naxis2, sizeof(int))) == NULL)
  { perror("\n\t calloc(ap)"); exit(EXIT_FAILURE); }

  for (j=0; j<naxis2; j++)
  {
    sprintf(ecor, "ECORD%d", j);
    if (get_FITS_key(hsize, header, ecor, val) == -1) ap[j]=j+par.ap1;
    else  sscanf(val, "%d", &ap[j]);
  }

  for (j=0; j<naxis2; j++)
  {
    if (par.vopt > 1) printf("\n*** aperture %d ***\n", ap[j]);
    memcpy(apdata, data[j], naxis1*sizeof(float));

    if ((imax=countlim(par, naxis1, ap[j], apdata, &ic1, &ic2)) == -1)
      continue;

    if (par.vopt > 1)
    {
      printf("accepted range: i=[%d:%d]  (%d points)\n",
              ic1+1, ic2+1, ic2-ic1+1);
      printf("----------\n");
    }

    id1=difmax(par, apdata, ic1, imax);
    id2=difmin(par, apdata, ic2, imax);

    if (par.vopt > 1)
    {
      printf("accepted range: [%d:%d] (%d points)\n", id1+1, id2+1, id2-id1+1);
      printf("----------\n");
    }

    if ((j == 0) || (j == naxis2-1))
    {
      sprintf(wstr1, "%16.10f", w1[j]+dw[j]*id1);
      sprintf(wstr2, "%16.10f", w1[j]+dw[j]*id2);
    }
    else
    {
      if (id1 == 0)       strcpy(wstr1, "INDEF");
      else                sprintf(wstr1, "%16.10f", w1[j]+dw[j]*id1);

      if (id2 == naxis1-1)  strcpy(wstr2, "INDEF");
      else                  sprintf(wstr2, "%16.10f", w1[j]+dw[j]*id2);
    }

    printf("%d  %s  %s  INDEF  INDEF\n", ap[j], wstr1, wstr2);
  }

  free(w1);
  free(dw);
  free(apdata);
  free(ap);
  for (i=0; i<naxis2; i++) free(data[i]);
  free(data);
  for (i=0; i<hsize; i++) free(header[i]);
  free(header);

  return(0);
}
/*** END ***/
