/*========================================================*/
/*                                                        */
/*  pfitsio.c       version 2.8.2         2007.01.25      */
/*                                                        */
/*  Copyright (C) 2007 by Wojtek Pych, CAMK PAN           */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*  Private FITS I/O library.                             */
/*                                                        */
/*  Definition of the FITS:                               */
/*  Hannish et al. 2001, A&A, 376, 359-380                */
/*                                                        */
/*========================================================*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pfitsio.h"

/*--------------------------------------------------------*/
/* Returns 1 on Intel architecture machines, 0 on Suns    */
/* Original function from DIA package by P. Wozniak       */
/*--------------------------------------------------------*/
int needswap(void)
{
        union { short s; char c[2]; } u;

  u.s=1;
  return((int)(u.c[0]));
}
/*--------------------------------------------------------*/
void swap_bytes(void *ptr, size_t ndata, int nbytes)
{
        char    c,
                *data;
        int     j, k;
        size_t  i, m;

  data=(char *)ptr;
  ndata*=nbytes;
  for (i=0; i<ndata; i+=nbytes)
  {
    k=nbytes/2;
    for (j=0; j<k; j++)
    {
      m=i+nbytes-1-j;
      c=data[i+j];
      data[i+j]=data[m];
      data[m]=c;
    }
  }

  return;
}
/*--------------------------------------------------------*/
/*  Read header card value with comment                   */
/*  Return card number (starting from 0)                  */
/*  or -1 on error.                                       */
/*--------------------------------------------------------*/
int get_FITS_key(int ncards, char **header, const char *keyword, char *value)
{
        char  lkeyword[KEYWORD_SIZE];
        int   i;              /*  loop numerator          */

  if (strlen(keyword) > KEYWORD_SIZE)
  {
    printf("WARNING! get_FITS_key(%s): invalid keyword\n", keyword);
    return(-1);
  }

  memcpy(lkeyword, keyword, strlen(keyword));
  for (i=(int)strlen(keyword); i<KEYWORD_SIZE; i++) lkeyword[i]=' ';

  for (i=0; i<ncards; i++)
  {
    if (strncmp(header[i], lkeyword, KEYWORD_SIZE) == 0)
    {
      memcpy(value, header[i]+10, VALUE_SIZE);  // value may start in column 11
      break;
    }
  }

  if (i == ncards) return(-1);

  return(i);
}
/*--------------------------------------------------------*/
/*  Return: image size in bytes (without the filling)     */
/*          0 no image                                    */
/*         -1 error                                       */
/*--------------------------------------------------------*/
size_t FITS_image_size(int ncards, char **header)
{
        char    val[VALUE_SIZE],
                keyword[KEYWORD_SIZE];
        int     i,
                p,
                bitpix,
                bytepix,
                naxis,
                naxisn;
        size_t  size;

  memset(val, (int)' ', VALUE_SIZE);

  if ((p=get_FITS_key(ncards, header, "BITPIX", val)) == -1)
  {
    printf("\n\tERROR! FITS_image_size(): BITPIX not found in the header\n");
    return(0);
  }
  if (p != 1)
    printf("WARNING! header does not conform to FITS standard (BITPIX)\n");
  if (sscanf(val, "%d", &bitpix) < 1)
  {
    printf("\n\tERROR! FITS_image_size(): reading BITPIX\n");
    return(0);
  }
  bytepix=abs(bitpix)/8;

  strcpy(keyword, "NAXIS");
  if ((p=get_FITS_key(ncards, header, keyword, val)) == -1)
  {
    printf("\n\tERROR! FITS_image_size(): NAXIS not found in the header\n");
    return(0);
  }
  if (p != 2)
    printf("WARNING! header does not conform to FITS standard (NAXIS)\n");

  if (sscanf(val, "%d", &naxis) < 1)
  {
    printf("\n\tERROR! FITS_image_size(): reading NAXIS\n");
    return(0);
  }

  if (naxis == 0) return(0);

  size=(size_t)bytepix;
  for (i=1; i<=naxis; i++)
  {
    if (snprintf((keyword+5), KEYWORD_SIZE-5, "%d", i) < 1)
    {
      printf("\n\tERROR! FITS_image_size(): %s\n", keyword);
      return(0);
    }
    if ((p=get_FITS_key(ncards, header, keyword, val)) == -1)
    {
      printf("\n\tERROR! FITS_image_size(): %s not found in the header\n",
              keyword);
      return(0);
    }
    if (p != i+2)
      printf("WARNING! header does not conform to FITS standard (%s)\n",
        keyword);

    if (sscanf(val, "%d", &naxisn) < 1)
    {
      printf("\n\tERROR! FITS_image_size(): reading %s\n", keyword);
      return(0);
    }
    size*=naxisn;
  }

  return(size);
}
/*--------------------------------------------------------*/
/*  Read FITS header (primary or extension)               */
/*  Return number of card images.                         */
/*--------------------------------------------------------*/
int read_FITS_header(FILE *inf, char ***header)
{
        char  **lheader;
        short eoh;            // end of header
        int   i,
              hc,             // number of header cards
              hr;             // number of header records

  eoh=0;

  if ((lheader=(char **)calloc(RECORD_CARDS, sizeof(char *))) == NULL)
  { perror("\n\t read_FITS_header(): calloc(lheader)"); exit(EXIT_FAILURE); }

  hc=0;
  for (hr=1; eoh==0; hr++)
  {
    if (!(lheader=(char **)realloc(lheader, hr*RECORD_CARDS*sizeof(char *))))
    { perror("\n\t read_FITS_header(): realloc(lheader)"); exit(EXIT_FAILURE); }

    for (i=0; i<RECORD_CARDS; i++)
    {
      if (!(lheader[hc+i]=(char *)calloc(CARD_SIZE, sizeof(char))))
      {
        perror("\n\t read_FITS_header(): calloc(lheader[hc+i])");
        exit(EXIT_FAILURE);
      }

      if (fread(lheader[hc+i], sizeof(char), CARD_SIZE, inf) != CARD_SIZE)
      {
        printf("\n\tERROR! read_FITS_header(): header corrupted\n");
        exit(EXIT_FAILURE);
      }
    }

    for (i=0; i<RECORD_CARDS; i++)
      if (strncmp(lheader[hc+i], "END     ", KEYWORD_SIZE) == 0) eoh=1;

    hc+=RECORD_CARDS;
  }

  *header=lheader;

  return(hc);
}
/*--------------------------------------------------------*/
/*  Read primary and extensions headers                   */
/*  Return number of extensions (0=primary header only)   */
/*  Byte offsets for extension headers recorded.          */
/*--------------------------------------------------------*/
int read_FITS_headers(FILE *inf, int **ncards, char ****header, long **offsets)
{
        char    ***lheader,         /* local pointer to the header  */
                value[VALUE_SIZE],  /* card value buffer            */
                val[VALUE_SIZE],    /* card value buffer            */
                tmp;
        short   eoF;
        int     p,
                hen,        /* header extension number (0=primary header)   */
                *lncards;   /* size of the header extension                 */
        long    offset,     /* FITS file offset                             */
                *loffsets;  /* table of FITS file offsets to the extensions */
        size_t  imsize;     /* image size                                   */

  memset(value, (int)' ', VALUE_SIZE);

/* read primary header */
  if ((lheader=(char ***)calloc(1, sizeof(char **))) == NULL)
  { perror("\n\t read_FITS_headers(): calloc(lheader)"); exit(EXIT_FAILURE); }

  if ((lncards=(int *)calloc(1, sizeof(int))) == NULL)
  { perror("\n\t read_FITS_headers(): calloc(lncards)"); exit(EXIT_FAILURE); }

  if ((loffsets=(long *)calloc(1, sizeof(long))) == NULL)
  { perror("\n\t read_FITS_headers(): calloc(loffsets)"); exit(EXIT_FAILURE); }

  loffsets[0]=0;
  if ((lncards[0]=read_FITS_header(inf, &lheader[0])) == 0)
  {
    free(lheader);
    free(lncards);
    free(loffsets);
    return(-1);
  }

/* check conformance with the FITS standard */
  if ((p=get_FITS_key(lncards[0], lheader[0], "SIMPLE", value)) == -1)
  {
    printf("\n\tERROR! read_FITS_headers(): SIMPLE not found in the header\n");
    free(lheader);
    free(lncards);
    free(loffsets);
    return(-1);
  }
  if (p != 0)
    printf("WARNING! Header does not conform to FITS standard (SIMPLE)\n");
  if (sscanf(value, "%s", val) < 1)
  {
    printf("\n\tERROR! read_FITS_headers(): reading SIMPLE\n");
    free(lheader);
    free(lncards);
    free(loffsets);
    return(-1);
  }
  if ((strcmp(val, "T") != 0) && (strncmp(val, "T/", 2) != 0))
    printf("WARNING! File does not conform to FITS standard (SIMPLE)\n");

/* check for possible extensions */
  if ((p=get_FITS_key(lncards[0], lheader[0], "EXTEND", value)) == -1)
  {
    *ncards=lncards;
    *header=lheader;
    *offsets=loffsets;

    return(0);
  }

  if (p < 3)
    printf("WARNING! Header does not conform to FITS standard (EXTEND)\n");
  if (sscanf(value, "%s", val) < 1)
  {
    printf("\n\tERROR! read_FITS_headers(): reading SIMPLE\n");
    free(lheader);
    free(lncards);
    free(loffsets);
    return(-1);
  }
  if ((strcmp(val, "T") != 0) && (strncmp(val, "T/", 2) != 0))
  {
    *ncards=lncards;
    *header=lheader;
    *offsets=loffsets;

    return(0);
  }

/* read extension headers */
  hen=0;
  eoF=0;
  while (eoF == 0)
  {
/* skip the image */
    imsize=FITS_image_size(lncards[hen], lheader[hen]);
    offset=(long)imsize;
    if (imsize%RECORD_SIZE != 0) offset+=(RECORD_SIZE-imsize%RECORD_SIZE);
    if (fseek(inf, offset, SEEK_CUR) != 0)
    {
      printf("\n\tERROR! read_FITS_headers(): fseek\n");
      free(lheader);
      free(lncards);
      free(loffsets);
      return(-1);
    }

/* check for the existence of another extension */
    imsize=fread(&tmp, 1, 1, inf);
    if (feof(inf) != 0) break;
    if (fseek(inf, -1, SEEK_CUR) != 0)
    {
      printf("\n\tERROR! read_FITS_headers(): fseek\n");
      free(lheader);
      free(lncards);
      free(loffsets);
      return(-1);
    }

/* read another header */
    hen++;

    if ((lheader=(char ***)realloc(lheader, (hen+1)*sizeof(char **))) == NULL)
    {
      perror("\n\t read_FITS_headers(): realloc(lheader)"); exit(EXIT_FAILURE);
    }

    if ((lncards=(int *)realloc(lncards, (hen+1)*sizeof(int))) == NULL)
    {
      perror("\n\t read_FITS_headers(): realloc(lncards)"); exit(EXIT_FAILURE);
    }

    if ((loffsets=(long *)realloc(loffsets, (hen+1)*sizeof(long))) == NULL)
    {
      perror("\n\t read_FITS_headers(): realloc(loffsets)"); exit(EXIT_FAILURE);
    }

    loffsets[hen]=ftell(inf);
    if ((lncards[hen]=read_FITS_header(inf, &lheader[hen])) == 0)
    {
      free(lheader);
      free(lncards);
      free(loffsets);
      return(-1);
    }

/* check the XTENSION value - only IMAGE supported */
    if ((p=get_FITS_key(lncards[hen], lheader[hen], "XTENSION", value)) == -1)
    {
      printf("\n\tERROR! read_FITS_headers(): XTENSION not found in the header\n");
      free(lheader);
      free(lncards);
      free(loffsets);
      return(-1);
    }
    if (p != 0)
      printf("WARNING! Header does not conform to FITS standard (XTENSION)\n");

    if (strncmp(value, "'IMAGE   '", 10) != 0)
      printf("WARNING! Unsupported extension type: %s\n", value);

    if (lncards[hen] == 0) eoF=1;
  }

  *ncards=lncards;
  *header=lheader;
  *offsets=loffsets;

  return(hen);
}
/*--------------------------------------------------------*/
int del_header_card(int ncards, char ***header, char *keyword)
{
        char  **lheader,
              val[VALUE_SIZE];
        int   p,
              i;

  lheader=*header;
  memset(val, (int)' ', VALUE_SIZE);

  if ((p=get_FITS_key(ncards, lheader, keyword, val)) == -1) return(ncards);

  for (i=p; i<ncards-1; i++)
    memcpy(lheader[i], lheader[i+1], CARD_SIZE);

  if ((p=get_FITS_key(ncards, lheader, "END", val)) == -1)
  {
    printf("\n\tERROR! del_header_card: header corrupted (END)\n");
    return(-1);
  }

  if (p%RECORD_CARDS != 0)
  {
    ncards-=RECORD_CARDS;

    for (i=0; i<RECORD_CARDS; i++) free(lheader[ncards+i]);
    if (!(lheader=(char **)realloc(lheader, ncards*sizeof(char *))))
    {
      perror("\n\t del_header_card(): realloc(lheader)"); exit(EXIT_FAILURE);
    }

    *header=lheader;
  }

  return(ncards);
}
/*--------------------------------------------------------*/
void write_FITS_header(FILE *outf, int ncards, char **header)
{
        char  tmp[CARD_SIZE];
        int   i;

  for (i=0; i<ncards; i++)
  {
    if (fwrite(header[i], 1, CARD_SIZE, outf) != CARD_SIZE)
    {
      printf("\n\tERROR! write_FITS_header: writing FITS header\n");
      exit(EXIT_FAILURE);
    }
  }

  if (ncards%RECORD_CARDS != 0)
  {
    memset(tmp, (int)' ', CARD_SIZE);
    for (i=0; i<RECORD_CARDS-ncards%RECORD_CARDS; i++)
    {
      if (fwrite(tmp, 1, CARD_SIZE, outf) != CARD_SIZE)
      {
        printf("\n\tERROR! write_FITS_header: writing FITS header\n");
        exit(EXIT_FAILURE);
      }
    }
  }

  return;
}
/*--------------------------------------------------------*/
size_t read_FITS_image(FILE *inf, char **header, int ncards, void **buf)
{
        void    *ptr;                     /* data table pointer   */
        char    keyword[KEYWORD_SIZE+1],  /* header entry keyword */
                value[VALUE_SIZE];
        int     i,              /* loop numerator               */
                naxes,          /* image dimension              */
                *naxis,         /* table of sizes of axes       */
                bitpix,         /* bits per pixel               */
                bytepix;        /* bytes per pixel              */
        size_t  pixnum;         /* number of image pixels       */

  memset(value, (int)' ', VALUE_SIZE);

  if (get_FITS_key(ncards, header, "NAXIS", value) == -1) return(0);
  if (sscanf(value, "%d", &naxes) < 1)
  {
    printf("\n\tERROR! read_FITS_image: reading NAXIS\n");
    return(0);
  }

  if ((naxis=(int *)calloc((size_t)naxes, sizeof(int))) == NULL)
  { perror("\n\t read_FITS_image: calloc(naxis)"); exit(EXIT_FAILURE); }

  for (i=0; i<naxes; i++)
  {
    if (snprintf(keyword, KEYWORD_SIZE, "NAXIS%d", i+1) < 6)
    {
      printf("\n\tERROR! read_FITS_image: setting NAXIS_ keyword\n");
      free(naxis);
      return(0);
    }
    if (get_FITS_key(ncards, header, keyword, value) == -1)
    {
      free(naxis);
      return(0);
    }
    if (sscanf(value, "%d", &naxis[i]) < 1)
    {
      printf("\n\tERROR! read_FITS_image: reading %s\n", keyword);
      free(naxis);
      return(0);
    }
  }

  pixnum=1;
  for (i=0; i<naxes; i++) pixnum*=naxis[i];

  free(naxis);

  if (get_FITS_key(ncards, header, "BITPIX", value) == -1) return(0);
  if (sscanf(value, "%d", &bitpix) < 1)
  {
    printf("\n\tERROR! read_FITS_image(): reading BITPIX\n");
    return(-1);
  }
  bytepix=abs(bitpix)/8;

  if ((ptr=(void *)malloc(pixnum*bytepix)) == NULL)
  { perror("\n\t read_FITS_image: malloc(ptr)"); exit(EXIT_FAILURE); }

  if (fread(ptr, (size_t)bytepix, pixnum, inf) != pixnum)
  { printf("\n\tERROR! reading FITS image\n"); exit(EXIT_FAILURE); }

  if (needswap() != 0) swap_bytes(ptr, pixnum, bytepix);

  *buf=ptr;

  return(pixnum);
}
/*--------------------------------------------------------*/
float *read_FITS_1Dfile(char *iname, char datatype,
                        int *ncards, char ***header, int *n1)
{
        void    *buf;
        char    value[VALUE_SIZE],
		**lhead;
        int     i,
		hs,
                bitpix,
                naxes,
                nx;
        size_t  nbytes;
        float   *ldata;
        double  FITS_bzero,
                FITS_bscale;
        FILE    *inf;

  memset(value, (int)' ', VALUE_SIZE);

  if ((inf=fopen(iname, "r")) == NULL)
  { printf("\n\t"); perror(iname); exit(EXIT_FAILURE); }

  hs=read_FITS_header(inf, &lhead);
  nbytes=read_FITS_image(inf, lhead, hs, &buf);

  if (fclose(inf) != 0)
    printf("\n\tWARNING! read_FITS_1Dfile: cannot close %s\n", iname);

  if (get_FITS_key(hs, lhead, "NAXIS", value) == -1)
  {
    printf("\n\tERROR! %s: NAXIS not found in the FITS header\n", iname);
    exit(EXIT_FAILURE);
  }
  if (sscanf(value, "%d", &naxes) < 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile: reading NAXIS\n"); exit(EXIT_FAILURE);
  }
  if (naxes != 1)
  {
    printf("\n\tERROR! NAXIS= %d: %s is not 1 dimensional\n", naxes, iname);
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hs, lhead, "NAXIS1", value) == -1)
  {
    printf("\n\tERROR! %s: NAXIS1 not found in the FITS header\n", iname);
    exit(EXIT_FAILURE);
  }
  if (sscanf(value, "%d", &nx) < 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile: reading NAXIS1\n"); exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hs, lhead, "BITPIX", value) == -1)
  {
    printf("\n\tERROR! %s: BITPIX not found in the FITS header\n", iname);
    exit(EXIT_FAILURE);
  }
  if (sscanf(value, "%d", &bitpix) < 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile: reading BITPIX\n"); exit(EXIT_FAILURE);
  }

  if ((ldata=(float *)calloc((size_t)nx, sizeof(long))) == NULL)
  { perror("\n\t read_FITS_1Dfile: calloc(ldata)"); exit(EXIT_FAILURE); }

  if (get_FITS_key(hs, lhead, "BZERO", value) == -1)  FITS_bzero=0.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bzero) < 1)
    {
      printf("\n\tERROR! read_FITS_1Dfile: reading BZERO\n");
      exit(EXIT_FAILURE);
    }
  }

  if (get_FITS_key(hs, lhead, "BSCALE", value) == -1) FITS_bscale=1.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bscale) < 1)
    {
      printf("\n\tERROR! read_FITS_1Dfile: reading BSCALE\n");
      exit(EXIT_FAILURE);
    }
  }

  switch(bitpix)
  {
    case   8: if (datatype == 'u')
                for (i=0; i<nx; i++)
                  ldata[i]=(float)((unsigned char *)buf)[i];
              else
                for (i=0; i<nx; i++)
                  ldata[i]=(float)((char *)buf)[i];
              break;
    case  16: if (datatype == 'u')
                for (i=0; i<nx; i++)
                  ldata[i]=(float)((unsigned short *)buf)[i];
              else
                for (i=0; i<nx; i++)
                  ldata[i]=(float)((short *)buf)[i];
              break;
    case  32: if (datatype == 'u')
                for (i=0; i<nx; i++)
                  ldata[i]=(float)((unsigned *)buf)[i];
              else
                for (i=0; i<nx; i++)
                  ldata[i]=(float)((int *)buf)[i];
              break;
    case -32: for (i=0; i<nx; i++)
                ldata[i]=((float *)buf)[i];
              break;
    case -64: for (i=0; i<nx; i++)
                ldata[i]=(float)((double *)buf)[i];
              break;
    default:  printf("\n\tERROR! %s: bitpix= %d not supported\n",
                      iname, bitpix);
              exit(EXIT_FAILURE);
  }

  free(buf);

  if (fabs(FITS_bscale-1.0) > EPSILON)
    for (i=0; i<nx; i++) ldata[i]*=FITS_bscale;
  if (fabs(FITS_bzero) > EPSILON)
    for (i=0; i<nx; i++) ldata[i]+=FITS_bzero;

  *n1=nx;
  *ncards=hs;
  *header=lhead;

  return(ldata);
}
/*--------------------------------------------------------*/
float **read_FITS_2Dfile(char *iname, char datatype,
                         int *ncards, char ***header, int *n1, int *n2)
{
        void    *buf;
        char    value[VALUE_SIZE],
		**lhead;
        int     i,
                j,
		hs,
                bitpix,
                naxes,
                nx,
                ny;
        size_t  nbytes;
        float   **ldata;
        double  FITS_bzero,
                FITS_bscale;
        FILE    *inf;

  memset(value, (int)' ', VALUE_SIZE);

  if ((inf=fopen(iname, "r")) == NULL)
  { printf("\n\t"); perror(iname); exit(EXIT_FAILURE); }

  hs=read_FITS_header(inf, &lhead);
  nbytes=read_FITS_image(inf, lhead, hs, &buf);

  if (fclose(inf) != 0)
    printf("\n\tWARNING! read_FITS_2Dfile: cannot close %s\n", iname);

  if (get_FITS_key(hs, lhead, "NAXIS", value) == -1)
  {
    printf("\n\tERROR! %s: NAXIS not found in the header\n", iname);
    exit(EXIT_FAILURE);
  }
  if (sscanf(value, "%d", &naxes) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile: reading NAXIS\n");
    exit(EXIT_FAILURE);
  }
  if (naxes != 2)
  {
    printf("\n\tERROR! NAXIS= %d: %s is not 2 dimensional\n", naxes, iname);
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hs, lhead, "NAXIS1", value) == -1)
  {
    printf("\n\tERROR! %s: NAXIS1 not found in header\n", iname);
    exit(EXIT_FAILURE);
  }
  if (sscanf(value, "%d", &nx) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile: reading NAXIS1\n");
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hs, lhead, "NAXIS2", value) == -1)
  {
    printf("\n\tERROR! %s: NAXIS2 not found in header\n", iname);
    exit(EXIT_FAILURE);
  }
  if (sscanf(value, "%d", &ny) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile: reading NAXIS2\n");
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hs, lhead, "BITPIX", value) == -1)
  {
    printf("\n\tERROR! %s: BITPIX not found in header\n", iname);
    exit(EXIT_FAILURE);
  }
  if (sscanf(value, "%d", &bitpix) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile: reading BITPIX\n");
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hs, lhead, "BZERO", value) == -1)  FITS_bzero=0.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bzero) < 1)
    {
      printf("\n\tERROR! read_FITS_2Dfile: reading BZERO\n");
      exit(EXIT_FAILURE);
    }
  }

  if (get_FITS_key(hs, lhead, "BSCALE", value) == -1) FITS_bscale=1.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bscale) < 1)
    {
      printf("\n\tERROR! read_FITS_2Dfile: reading BSCALE\n");
      exit(EXIT_FAILURE);
    }
  }

  if ((ldata=(float **)calloc((size_t)ny, sizeof(float *))) == NULL)
  { perror("\n\t read_FITS_2Dfile: calloc(ldata)"); exit(EXIT_FAILURE); }
  for (i=0; i<ny; i++)
  {
    if ((ldata[i]=(float *)calloc((size_t)nx, sizeof(float))) == NULL)
    {
      perror("\n\t read_FITS_2Dfile: calloc(ldata[i])");
      exit(EXIT_FAILURE);
    }
  }

  switch(bitpix)
  {
    case   8: if (datatype == 'u')
                for (i=0; i<ny; i++)
                  for (j=0; j<nx; j++)
                    ldata[i][j]=(float)((unsigned char *)buf)[i*nx+j];
              else
                for (i=0; i<ny; i++)
                  for (j=0; j<nx; j++)
                    ldata[i][j]=(float)((char *)buf)[i*nx+j];
              break;
    case  16: if (datatype == 'u')
                for (i=0; i<ny; i++)
                  for (j=0; j<nx; j++)
                    ldata[i][j]=(float)((unsigned short *)buf)[i*nx+j];
              else
                for (i=0; i<ny; i++)
                  for (j=0; j<nx; j++)
                    ldata[i][j]=(float)((short *)buf)[i*nx+j];
              break;
    case  32: if (datatype == 'u')
                for (i=0; i<ny; i++)
                  for (j=0; j<nx; j++)
                    ldata[i][j]=(float)((unsigned *)buf)[i*nx+j];
              else
                for (i=0; i<ny; i++)
                  for (j=0; j<nx; j++)
                    ldata[i][j]=(float)((int *)buf)[i*nx+j];
              break;
    case -32: for (i=0; i<ny; i++)
                for (j=0; j<nx; j++)
                  ldata[i][j]=((float *)buf)[i*nx+j];
              break;
    case -64: for (i=0; i<ny; i++)
                for (j=0; j<nx; j++)
                  ldata[i][j]=(float)((double *)buf)[i*nx+j];
              break;
    default:  printf("\n\tERROR! %s: bitpix=%d not supported\n", iname, bitpix);
              exit(EXIT_FAILURE);
  }

  free(buf);

  if (fabs(FITS_bscale-1.0) > EPSILON)
    for (i=0; i<ny; i++)
      for (j=0; j<nx; j++)
        ldata[i][j]*=FITS_bscale;

  if (fabs(FITS_bzero) > EPSILON)
    for (i=0; i<ny; i++)
      for (j=0; j<nx; j++)
        ldata[i][j]+=FITS_bzero;

  *n1=nx;
  *n2=ny;
  *header=lhead;
  *ncards=hs;

  return(ldata);
}
/*--------------------------------------------------------*/
void scale(int size, float *data, short *idata)
{
        int     i;
        float   min,
                max;
        double  FITS_bscale,
                FITS_bzero;

/** scalling factors **/
  min = max = data[0];
  for (i=1; i<size; i++)
  {
    if (data[i] > max) max=data[i];
    if (data[i] < min) min=data[i];
  }
  FITS_bzero=(max-min)/2.0;
  FITS_bscale=(max-min)/65535.0;

  for (i=0; i<size; i++)
    idata[i]=(short)rint((data[i]-FITS_bzero)/FITS_bscale);

  return;
}
/*--------------------------------------------------------*/
void write_FITS_2Dimage(FILE *outf, int nx, int ny, int bytepix, void **data)
{
        char    *tmp;
        int     i;
        size_t  size,   // image size in bytes
                ls;

  if (needswap() != 0)
  {
    for (i=0; i<ny; i++)
    {
      swap_bytes(data[i], (size_t)nx, bytepix);
      if (fwrite(data[i], (size_t)bytepix, (size_t)nx, outf) != (size_t) nx)
      {
        perror("\n\t write_FITS_2Dimage: fwrite(data[i])");
        exit(EXIT_FAILURE);
      }
    }
  }
  else
    for (i=0; i<ny; i++)
      if (fwrite(data[i], (size_t)bytepix, (size_t)nx, outf) != (size_t)nx)
      {
        perror("\n\t write_FITS_2Dimage: fwrite(data[i])");
        exit(EXIT_FAILURE);
      }

  size=(size_t)nx*ny*bytepix;
  if ((ls=size%RECORD_SIZE) != 0) ls=RECORD_SIZE-ls;
  if (!(tmp=(char *)malloc(ls)))
  {
    perror("\n\t write_FITS_2Dimage(): malloc(tmp)");
    exit(EXIT_FAILURE);
  }
  memset(tmp, 0, ls);
  if (fwrite(tmp, 1, ls, outf) != ls)
  {
    printf("\n\tERROR! write_FITS_2Dimage: writing zero padding to the file\n");
    exit(EXIT_FAILURE);
  }
  free(tmp);

  return;
}
/*--------------------------------------------------------*/
void write_FITS_2Dfile(char *oname, int ncards, char **header, int nx, int ny,
                       int bytepix, void **data)
{
        FILE  *outf;

  if (!(outf=fopen(oname, "w")))
  {
    printf("\n\t");
    perror(oname);
    exit(EXIT_FAILURE);
  }

  write_FITS_header(outf, ncards, header);
  write_FITS_2Dimage(outf, nx, ny, bytepix, data);

  if (fclose(outf) != 0)
    printf("\n\tERROR! write_FITS_2Dfile: cannot close %s\n", oname);

  return;
}

/*** END ***/
